@@ -10,6 +10,7 @@ import requests
 import io
 import logging
 from .md import parse_markdown
+from tornado.httpclient import HTTPRequest
 
 class MaxAPI:
     """
@@ -34,10 +35,12 @@ class MaxAPI:
     def __init__(self, auth_token: str = None, on_event=None):
         """
         Initializes the MaxAPI instance.
-        This constructor will block until the connection is established and authenticated.
         
         Args:
-            auth_token (str): The authentication token for the session.
+            auth_token (str, optional): The authentication token for the session. If provided,
+                                        the API will connect and authenticate automatically.
+                                        If not provided, the API will connect and wait for
+                                        manual authentication via send_vertify_code and check_vertify_code.
             on_event (callable, optional): A callback function to handle server-push events.
                                            It receives one argument: the event data dictionary.
         """
@@ -51,6 +54,7 @@ class MaxAPI:
         }
         self.user = None
         self.chats = {}
+        self.subscribed_chats = set()
 
         self.ws = None
         self.ioloop = None
@@ -73,13 +77,11 @@ class MaxAPI:
 
         self._start_ioloop()
         
+        # We wait for the connection to be established, but not necessarily authenticated.
         is_ready = self.ready_event.wait(timeout=20)
-        if not is_ready or not self.is_running:
-            failure_reason = "Failed to connect and authenticate within the timeout period."
-            if not self.is_running:
-                failure_reason = "Connection failed during initialization."
+        if not is_ready:
             self.close()
-            raise TimeoutError(failure_reason)
+            raise TimeoutError("Failed to connect to WebSocket within the timeout period.")
 
     def _signal_handler(self, signum, frame):
         self.logger.info(f"\nSignal {signum} received, initiating shutdown...")
@@ -103,20 +105,38 @@ class MaxAPI:
 
     @tornado.gen.coroutine
     def _connect_and_run(self):
-        """Main async task: connects, launches listener, authenticates, and signals readiness."""
+        """Main async task: connects, launches listener, authenticates (if token exists), and signals readiness."""
         while True:
             try:
-                self.logger.info("Connecting...")
-                self.ws = yield tornado.websocket.websocket_connect(self.ws_url)
+                self.logger.info('Connecting...')
+                # Create a custom HTTPRequest with Origin header
+                request = HTTPRequest(
+                    url=self.ws_url,
+                    headers={
+                        "Origin": "https://web.max.ru",
+                        "User-Agent": self.user_agent["headerUserAgent"],
+                        "Sec-Fetch-Dest": "empty",
+                        "Sec-Fetch-Mode": "websocket",
+                        "Sec-Fetch-Site": "cross-site",
+                    }
+                )
+                self.ws = yield tornado.websocket.websocket_connect(request)
                 self.is_running = True
                 self.logger.info("Connected to WebSocket.")
                 self.ioloop.add_callback(self._listener_loop_async)
                 yield self._handshake_async()
-                yield self._authenticate_async()
+                
+                # Only authenticate if a token was provided during initialization
+                if self.token:
+                    yield self._authenticate_async()
+                    self.logger.info("API is online and ready.")
+                else:
+                    self.logger.info("API is connected. Please authenticate using verification code methods.")
 
-                self.heartbeat_callback = tornado.ioloop.PeriodicCallback(self._send_heartbeat, 5000)
+                self.heartbeat_callback = tornado.ioloop.PeriodicCallback(self._send_heartbeat, 3000)
                 self.heartbeat_callback.start()
-                self.logger.info("API is online and ready.")
+                
+                # Signal that the connection is ready for commands
                 self.ready_event.set()
                 break
             except Exception as e:
@@ -125,21 +145,88 @@ class MaxAPI:
         
     @tornado.gen.coroutine
     def _listener_loop_async(self):
-        """Asynchronously listens for all incoming messages."""
+        """Asynchronously listens for all incoming messages. Auto-reconnects on server close."""
         try:
             while self.is_running:
                 message = yield self.ws.read_message()
                 if message is None:
                     if self.is_running:
-                        self.logger.warning("Connection closed by server.")
+                        self.logger.warning("Connection closed by server. Attempting to reconnect...")
+                        # Trigger reconnect in async context
+                        yield self._reconnect_async()
                     break
                 self._process_message(message)
         except tornado.websocket.WebSocketClosedError:
-             if self.is_running: self.logger.warning("Listener loop terminated: WebSocket closed.")
+            if self.is_running:
+                self.logger.warning("Listener loop terminated: WebSocket closed by server. Reconnecting...")
+                yield self._reconnect_async()
         except Exception as e:
-            if self.is_running: self.logger.error(f"An error occurred in the listener loop: {e}")
+            if self.is_running:
+                self.logger.error(f"An error occurred in the listener loop: {e}")
+                yield self._reconnect_async()
         finally:
             self.is_running = False
+
+    @tornado.gen.coroutine
+    def _reconnect_async(self):
+        """Async reconnect logic triggered internally after connection loss."""
+        self.is_running = False
+        if self.heartbeat_callback:
+            self.heartbeat_callback.stop()
+
+        # Close existing WS if still open
+        if self.ws:
+            try:
+                self.ws.close()
+            except:
+                pass
+
+        # Wait a moment before reconnecting
+        yield tornado.gen.sleep(2)
+
+        # Re-initialize connection
+        try:
+            self.logger.info('Reconnecting...')
+            request = HTTPRequest(
+                url=self.ws_url,
+                headers={
+                    "Origin": "https://web.max.ru",
+                    "User-Agent": self.user_agent["headerUserAgent"],
+                    "Sec-Fetch-Dest": "empty",
+                    "Sec-Fetch-Mode": "websocket",
+                    "Sec-Fetch-Site": "cross-site",
+                }
+            )
+            self.ws = yield tornado.websocket.websocket_connect(request)
+            self.is_running = True
+            self.logger.info("Reconnected to WebSocket.")
+            self.ioloop.add_callback(self._listener_loop_async)
+            yield self._handshake_async()
+
+            # Re-authenticate if token exists
+            if self.token:
+                yield self._authenticate_async()
+                self.logger.info("Re-authenticated successfully.")
+
+            # Restart heartbeat
+            if self.heartbeat_callback:
+                self.heartbeat_callback.start()
+
+            for chat_id in self.subscribed_chats:
+                try:
+                    yield self.send_command_async(self.OPCODE_MAP['SUBSCRIBE_TO_CHAT'], {"chatId": chat_id, "subscribe": True})
+                    self.logger.debug(f"Resubscribed to chat {chat_id}")
+                except Exception as e:
+                    self.logger.warning(f"Failed to resubscribe to chat {chat_id}: {e}")
+
+            self.logger.info("Reconnection and re-authentication complete.")
+        except Exception as e:
+            self.logger.error(f"Reconnection failed: {e}. Will retry in 5 seconds...")
+            yield tornado.gen.sleep(5)
+            # Try again recursively (limited by caller's logic or external control)
+            # Or you can raise and let parent handle
+            if self.is_running:
+                yield self._reconnect_async()  # Recursive retry
     
     def _process_message(self, message):
         """Processes a raw message, dispatching to sync/async waiters or event handlers."""
@@ -152,6 +239,10 @@ class MaxAPI:
                     pending_request = self.pending_responses.get(seq_id)
                 
                 if pending_request:
+                    original_opcode = pending_request.get("opcode")
+                    if original_opcode != self.OPCODE_MAP['HEARTBEAT']:
+                        self.logger.debug(f"API Response (for Opcode {original_opcode}, Seq {seq_id}): {json.dumps(data, ensure_ascii=False)}")
+
                     if "event" in pending_request:
                         pending_request["response"] = data
                         pending_request["event"].set()
@@ -159,17 +250,20 @@ class MaxAPI:
                         with self.response_lock:
                             self.pending_responses.pop(seq_id, None)
                         pending_request["future"].set_result(data)
-            else:
+            elif data.get("cmd") == 0:
                 # This is a server-push event, not a response to a command.
                 if self.on_event:
                     self.ioloop.run_in_executor(None, self.on_event, data)
+            else:
+                self.logger.debug(f"Received API error(?): {json.dumps(data, indent=4, ensure_ascii=False)}")
+        
         except Exception as e:
             self.logger.error(f"Error processing message: {e}")
 
     def close(self):
         if not self.is_running and self.ioloop is None: return
         self.logger.info("Closing connection...")
-        self.is_running = False
+        self.is_running = False  # Prevent reconnect attempts
 
         if self.ioloop:
             self.ioloop.add_callback(self._shutdown_async)
@@ -200,18 +294,16 @@ class MaxAPI:
         
         future = tornado.gen.Future()
         with self.response_lock:
-            self.pending_responses[seq_id] = {"future": future}
+            self.pending_responses[seq_id] = {"future": future, "opcode": opcode}
 
         try:
             yield self.ws.write_message(json.dumps(command))
-            # The async future handler in _process_message already pops, so this is clean.
             response = yield tornado.gen.with_timeout(
                 self.ioloop.time() + timeout,
                 future
             )
             raise tornado.gen.Return(response)
         except tornado.gen.TimeoutError:
-            # Cleanup on timeout
             with self.response_lock:
                 self.pending_responses.pop(seq_id, None)
             raise TimeoutError(f"Async request (opcode: {opcode}, seq: {seq_id}) timed out.")
@@ -226,6 +318,10 @@ class MaxAPI:
     @tornado.gen.coroutine
     def _authenticate_async(self):
         self.logger.info("Authenticating...")
+        if not self.token:
+            self.logger.error("Authentication failed: No token available.")
+            return
+
         payload = {
             "interactive": True, "token": self.token,
             "chatsSync": 0, "contactsSync": 0, "presenceSync": 0,
@@ -234,7 +330,7 @@ class MaxAPI:
         response = yield self.send_command_async(self.OPCODE_MAP['AUTHENTICATE'], payload)
         response = response['payload']
         self.logger.info(f"Authentication successful. User: {response['profile']['contact']['names'][0]['name']}")
-        self.user = response['profile']
+        self.user = response['profile']['contact']
         chats = {}
         for item in response['chats']:
             item_id = str(item.get('id'))
@@ -247,13 +343,11 @@ class MaxAPI:
     def _send_heartbeat(self):
         if not self.is_running: return
         try:
-            # Heartbeat doesn't need a response, so wait_for_response=False is correct
             self.send_command(self.OPCODE_MAP['HEARTBEAT'], {"interactive": False}, wait_for_response=False)
         except tornado.websocket.WebSocketClosedError:
             self.logger.warning("Heartbeat failed: WebSocket is closed.")
             self.is_running = False
         except Exception as e:
-            # Catch other potential errors during send
             if self.is_running:
                 self.logger.error(f"Heartbeat failed with error: {e}")
                 self.is_running = False
@@ -276,7 +370,7 @@ class MaxAPI:
                     return None
                 event = threading.Event()
                 with self.response_lock:
-                    self.pending_responses[seq_id] = {"event": event, "response": None}
+                    self.pending_responses[seq_id] = {"event": event, "response": None, "opcode": opcode}
                 self.ioloop.add_callback(self.ws.write_message, json.dumps(command))
                 is_set = event.wait(timeout)
                 with self.response_lock:
@@ -291,14 +385,12 @@ class MaxAPI:
             except Exception:
                 if attempt == max_attempts:
                     raise
-                # Small delay before retrying
                 time.sleep(1)
 
     def _reconnect(self):
         self.logger.info("Attempting to reconnect to WebSocket...")
         self.close()
         self._start_ioloop()
-        # Wait for re-initialization
         if not self.ready_event.wait(timeout=10):
             self.is_running = False
             raise ConnectionError("Failed to reconnect after multiple attempts.")
@@ -334,6 +426,10 @@ class MaxAPI:
         status = "Subscribed to" if subscribe else "Unsubscribed from"
         response = self.send_command(self.OPCODE_MAP['SUBSCRIBE_TO_CHAT'], payload)
         self.logger.info(f"{status} chat {chat_id}")
+        if subscribe:
+            self.subscribed_chats.add(chat_id)
+        else:
+            self.subscribed_chats.discard(chat_id)
         return response
 
     def mark_as_read(self, chat_id: int, message_id: str):
@@ -355,20 +451,59 @@ class MaxAPI:
         return self.chats
     
     def send_vertify_code(self, phone_number: str):
+        """Sends a verification code to a phone number to start the authentication process."""
         payload = {
             "phone": phone_number,
             "type": "START_AUTH",
             "language": "ru"
 	    }
-        return self.send_command(self.OPCODE_MAP['SEND_VERTIFY_CODE'], payload)
+        res = self.send_command(self.OPCODE_MAP['SEND_VERTIFY_CODE'], payload)
+        # Temporarily store the token received for the code verification step
+        if 'payload' in res and 'token' in res['payload']:
+            self.token = res['payload']['token']
+            self.logger.info('Saved temp auth-token')
+        return res
     
-    def check_vertify_code(self, code: int):
+    def check_vertify_code(self, code: str):
+        """Checks the verification code and completes authentication."""
+        if not self.token:
+            raise RuntimeError("Cannot check verification code. Please call send_vertify_code first.")
+        
         payload = {
             "token": self.token,
             "verifyCode": code,
             "authTokenType": "CHECK_CODE"
         }
-        return self.send_command(self.OPCODE_MAP['CHECK_VERTIFY_CODE'], payload)
+        res = self.send_command(self.OPCODE_MAP['CHECK_VERTIFY_CODE'], payload, wait_for_response=True)
+        
+        # On success, a new, permanent token is issued.
+        token = res['payload'].get('tokenAttrs', {}).get('LOGIN', {}).get('token')
+        if token:
+            self.token = token
+            # Now that we have the permanent token, trigger the full authentication
+            self.logger.info("Verification successful. Finalizing authentication...")
+
+            # Instead of run_sync, use proper async coordination
+            auth_event = threading.Event()
+            auth_result = [None]  # [exception or None]
+
+            def _run_authenticate():
+                try:
+                    future = self._authenticate_async()
+                    self.ioloop.add_future(future, lambda f: auth_event.set())
+                except Exception as e:
+                    auth_result[0] = e
+                    auth_event.set()
+
+            self.ioloop.add_callback(_run_authenticate)
+            auth_event.wait()
+            if auth_result[0] is not None:
+                raise auth_result[0]
+            
+            self.logger.info("API is online and ready.")
+            return self.token
+            
+        return res
     
     def send_generic_command(self, command_name: str, payload: dict, wait_for_response: bool = True, timeout: int = 10):
         command_name_upper = command_name.upper()
